"""Top level interfaces and definitions for Lakshmi."""

import yaml

import lakshmi.utils as utils
from lakshmi.assets import from_dict, to_dict
from lakshmi.table import Table


class Account:
    """Class representing an account (collection of assets)."""

    def __init__(self, name, account_type):
        """
        Args:
          name: Printable name for this account.
          account_type: Type of this account (string).
        """
        self._name = name
        self.account_type = account_type
        self._assets = {}
        self._cash = 0

    def to_dict(self):
        """Returns a dict representing this object."""
        d = {'Name': self._name,
             'Account Type': self.account_type}
        if self._assets:
            d['Assets'] = [to_dict(asset) for asset in self._assets.values()]
        if self._cash != 0:
            d['Available Cash'] = self._cash
        return d

    @classmethod
    def from_dict(cls, d):
        """Returns a new object specified by dictionary d.

        This is reverse of to_dict.
        Args:
            d: A dictionary (usually the output of to_dict).

        Returns: A new Account object.

        Raises: AssertionError if d cannot be parsed correctly.
        """
        ret_obj = Account(d.pop('Name'), d.pop('Account Type'))
        for asset_dict in d.pop('Assets', []):
            ret_obj.add_asset(from_dict(asset_dict))
        ret_obj._cash = d.pop('Available Cash', 0)
        assert len(d) == 0, 'Extra attributes found: ' + str(list(d.keys()))
        return ret_obj

    def string(self):
        """Returns a string representation of this object."""
        table = Table(2)
        table.add_row(['Name:', f'{self._name}'])
        table.add_row(['Type:', f'{self.account_type}'])
        total = sum([asset.adjusted_value()
                    for asset in self._assets.values()])
        table.add_row(['Total:', utils.format_money(total)])
        if self._cash:
            table.add_row(['Available Cash:',
                           utils.format_money_delta(self._cash)])
        return table.string(tablefmt='plain')

    def add_asset(self, asset, replace=False):
        """Add an asset to this Account.

        Args:
            asset: A lakshmi.assets.Asset object. The short_name of the
            asset must be unique in this account (unless replace is set
            to True).
            replace: If true, this asset will replace any other asset
            with the same short_name.

        Returns: The new account.

        Raises: AssertionError if another asset with the same short_name
        is present in the account (unless replace is True).
        """
        assert replace or asset.short_name() not in self._assets, (
            f'Attempting to add duplicate Asset: {asset.short_name()}')
        self._assets[asset.short_name()] = asset
        return self

    def assets(self):
        """Retuns all assets in this account."""
        return self._assets.values()

    def set_assets(self, assets):
        """Replaces all assets in this account with provided assets.

        Args:
            assets: A list of lakshmi.assets.Asset
        """
        self._assets = {}
        for asset in assets:
            self.add_asset(asset)

    def get_asset(self, short_name):
        """Returns an asset specified by short_name."""
        return self._assets[short_name]

    def remove_asset(self, short_name):
        """Removes the asset in this account specified by short_name."""
        del self._assets[short_name]

    def name(self):
        """Returns the name of this account."""
        return self._name

    def add_cash(self, delta):
        """Adds cash to this account."""
        self._cash += delta

    def available_cash(self):
        """Returns the available cash in this account."""
        return self._cash


class AssetClass:
    """This class represents (a tree of) Asset classes.

    Each asset class is specified by a name and can optionally have child
    asset classes. The child asset classes have a ratio (float) associated
    with them and the sum of all ratios must add to 1.0.

    Example of asset classes: Stocks, Bonds, US, International, Small Cap,
    Large Cap, etc.
    Some asset classes can have children asset classes. E.g.
    Stock -> US (60%) and International (40%).
    """

    def __init__(self, name):
        """Returns a new AssetClass object named name."""
        self.name = name
        self._children = []
        # Populated when _validate is called.
        self._leaves = None

    def children(self):
        """Returns a list of all the children of this asset class."""
        return self._children

    def to_dict(self):
        """Returns a dict representing this object."""
        d = {'Name': self.name}
        if self._children:
            d['Children'] = [{'Ratio': ratio} | child.to_dict()
                             for child, ratio in self._children]
        return d

    @classmethod
    def from_dict(cls, d):
        """Returns a new object specified by dictionary d.

        This is reverse of to_dict.
        Args:
            d: A dictionary (usually the output of to_dict).

        Returns: A new AssetClass object.

        Raises: AssertionError if d cannot be parsed correctly.
        """
        ret_obj = AssetClass(d.pop('Name'))
        for child_dict in d.pop('Children', []):
            ret_obj.add_subclass(
                child_dict.pop('Ratio'),
                AssetClass.from_dict(child_dict))
        assert len(d) == 0, f'Extra attributes found: {list(d.keys())}'
        return ret_obj.validate()

    def copy(self):
        """Returns a copy of this AssetClass and its sub-classes."""
        ret_val = AssetClass(self.name)
        for child, ratio in self._children:
            ret_val.add_subclass(ratio, child.copy())
        return ret_val

    def add_subclass(self, ratio, asset_class):
        """Add a subclass (asset class) to this asset class.

        Args:
            ratio: The ratio of this asset class.
            asset_class: An AssetClass object representing the child class.
        """
        self._children.append((asset_class, ratio))
        # leaves is not upto date now, need validation again.
        self._leaves = None
        return self

    def _validate(self):
        """Internal helper method to validate this asset class.

        Returns: A tuple (leaf names, class names) for the subtree containing
        set of all the leaf asset class names and list of all (including
        non-leaf) asset classes.

        Raises: AssertionError If ratio is not a valid float in [0, 1] or if
        the sum of ratio of all the sub-classes doesn't add up to 1.0.
        """
        if not self._children:
            self._leaves = {self.name}
            return self._leaves, [self.name]

        self._leaves = set()
        class_names = [self.name]
        total = 0.0
        for asset_class, ratio in self._children:
            assert ratio >= 0.0 and ratio <= 1.0, (
                f'Bad ratio provided to Asset Class ({ratio})')
            total += ratio
            temp_leafs, temp_classes = asset_class._validate()
            self._leaves.update(temp_leafs)
            class_names += temp_classes

        # Check if all percentages add up to 100%
        assert abs(total - 1) < 1e-6, (
            f'Sum of sub-classes is not 100% (actual: {total * 100}%)')

        return self._leaves, class_names

    def validate(self):
        """Validates if this asset class tree has any errors.

        This function checks the ratios and also if the asset class tree has
        any duplicate asset class names. It also sets an internal field which
        is used by few other methods in this class. Those methods will raise
        an exception if they are called without calling validate.

        Returns: self

        Raises: AssertionError if there are any duplicate asset classes in the
        tree.
        """
        unused_leaves, all_class_names = self._validate()
        duplicates = set([x for x in all_class_names
                          if all_class_names.count(x) > 1])

        assert not duplicates, f'Found duplicate Asset class(es): {duplicates}'
        return self

    def _check(self):
        """Internal method to check if validate has been called."""
        assert self._leaves, (
            'Need to validate AssetAllocation before using it.')

    def find_asset_class(self, asset_class_name):
        """Returns asset_class_name object and its desired ratio.

        validate() must be called before calling this method.
        This methods searches for an asset class object with name
        asset_class_name in the current asset tree and returns the found
        object and the effective _absolute_ ratio of that asset class as
        a tuple. If the asset_class_name is not found, this method returns
        None. For example, if the asset class tree is:
        All -> Stocks 60%, Bonds 40%
        Stocks -> US 60%, Intl 40%
        find_asset_class('US') will return asset class representing the 'US'
        node and 0.36 as ratio (= 0.6 * 0.6).

        Args:
            asset_class_name: A string representing the name to be searched.

        Returns: (Found asset class object, absolute ratio (float)) or None
        if the asset_class_name is not found in this asset class tree.

        Raises: AssertionError if validate is not callled before calling this
        method.
        """
        self._check()

        if self.name == asset_class_name:
            return self, 1.0

        for asset_class, ratio in self._children:
            ret_value = asset_class.find_asset_class(asset_class_name)
            if ret_value:
                return ret_value[0], ret_value[1] * ratio

        return None

    def leaves(self):
        """Returns all leaf asset class names.

        validate() must be called before calling this method.

        Returns: A list of leaf asset class names.

        Raises: AssertionError if validate is not callled before calling this
        method.
        """
        self._check()
        return self._leaves

    def value_mapped(self, money_allocation):
        """Returns how much money is mapped to this Asset Class.

        validate() must be called before calling this method.
        Given a money allocation, this class return the amount of money mapped
        to this asset class or its childen.

        Ars:
          money_allocation: A map of leaf_class_names (string) -> money
          (float).

        Returns: Total amount of money mapped to this asset class.

        Raises: AssertionError if validate is not callled before calling this
        method.
        """
        self._check()
        return sum([value for name, value in money_allocation.items()
                    if name in self._leaves])

    class Allocation:
        """This class is a convenience class to represent the return value of
        return_allocation method. This class represents a partcular node of
        AssetClass + its direct children and also the money allocated to the
        asset class and its direct children. It is meant to be used as a
        data-only class.
        """
        class Children:
            """Class representing a child of Allocation class. This class is
            meant to be used as a data-only class.
            """
            def __init__(self, name, actual_allocation, desired_allocation,
                         value, value_difference):
                """
                Args:
                    name: Name of the child asset class.
                    actual_allocation: Actual allocation (ratio, float).
                    desired_allocation: Desired asset allocation (ratio,
                    float).
                    value: The absolute amount of money allocated to this
                    child.
                    value_difference: The difference between desired and
                    actual allocation.
                """
                self.name = name
                self.actual_allocation = actual_allocation
                self.desired_allocation = desired_allocation
                self.value = value
                self.value_difference = value_difference

        def __init__(self, name, value):
            """
            Args:
                name: Name of this asset class.
                value: The absolute amount of money allocated to this asset
                class.
            """
            self.name = name
            self.value = value
            self.children = []

        def add_child(self, name, actual, desired):
            """Add a child asset to this asset class.

            Args:
                name: Name of the child asset class.
                actual: The actual amount of money allocated to this child.
                desired: The desired alloction of money for this child.
            """
            self.children.append(
                self.Children(name, actual, desired, actual * self.value,
                              (desired - actual) * self.value))

    def return_allocation(self, money_allocation, levels=-1):
        """Returns actual and desired allocation based on how money is allocated.

        Args:
          money_allocation: A map of leaf_class_name -> money.
          levels: How many levels of child allocation to return (-1 = all).

        Returns: A list of Allocation objects (for itself and any child classes
        based on the levels flag).
        """
        value = self.value_mapped(money_allocation)
        actual_alloc = self.Allocation(self.name, value)

        if value == 0:
            return [actual_alloc]

        for asset_class, desired_ratio in self._children:
            actual_ratio = asset_class.value_mapped(money_allocation) / value
            actual_alloc.add_child(
                asset_class.name,
                actual_ratio,
                desired_ratio)

        ret_val = [actual_alloc]

        if levels == 0:
            return ret_val

        if levels > 0:
            levels -= 1

        for asset_class, unused_ratio in self._children:
            ret_val += asset_class.return_allocation(money_allocation, levels)

        return ret_val


class Portfolio:
    """Top-level class representing a portfolio.

    A portfolio consists of multiple accounts (Account objects). This
    class provides helper methods to add/delete accounts etc. This class
    also provides methods for portfolio level operations.
    """
    def __init__(self, asset_classes):
        """
        Args:
            asset_classes: An AssetClass object representing the desired asset
            allocation.
        """
        self._accounts = {}
        self.asset_classes = asset_classes.validate()
        self._leaf_asset_classes = asset_classes.leaves()

    def save(self, filename):
        """Save this portfolio to a file."""
        with open(filename, 'w') as f:
            yaml.dump(self.to_dict(), f, sort_keys=False)

    @classmethod
    def load(cls, filename):
        """Loads and returns portfolio from file."""
        with open(filename) as f:
            d = yaml.load(f.read(), Loader=yaml.SafeLoader)
        return Portfolio.from_dict(d)

    def to_dict(self):
        """Returns a dictionary representation of this portfolio."""
        d = {'Asset Classes': self.asset_classes.to_dict()}
        if self._accounts:
            d['Accounts'] = [account.to_dict()
                             for account in self._accounts.values()]
        return d

    @classmethod
    def from_dict(cls, d):
        """Returns a portfolio represented by dict d (reverse of to_dict)."""
        ret_obj = Portfolio(AssetClass.from_dict(d.pop('Asset Classes')))
        for account_dict in d.pop('Accounts', []):
            ret_obj.add_account(Account.from_dict(account_dict))
        assert len(d) == 0, f'Extra attributes found: {list(d.keys())}'
        return ret_obj

    def add_account(self, account, replace=False):
        """Add an account to this portfolio.

        This method adds a new account to this portfolio. The account
        name must be unique unless replace is set to True.

        Returns: self object with account added.

        Raises:
            AssertionError: If any of the assets are mapped to asset classes
            that doesn't match the asset classes given the asset allocation
            provided when creating this class.
            AssertError: If another account with the same exists in the
            portfolio and replace=False.
        """
        for asset in account.assets():
            for asset_class in asset.class2ratio.keys():
                assert asset_class in self._leaf_asset_classes, (
                    f'Unknown or non-leaf asset class: {asset_class}')

        assert replace or account.name() not in self._accounts, (
            f'Attempting to add duplicate account: {account.name()}')

        self._accounts[account.name()] = account
        return self

    def remove_account(self, account_name):
        """Delete account specifed by account_name."""
        del self._accounts[account_name]

    def accounts(self):
        """Returns a list (dict_values) of accounts."""
        return self._accounts.values()

    def get_account(self, name):
        """Return Account object represented by name."""
        return self._accounts[name]

    def get_account_name_by_substr(self, account_str):
        """Returns account name whose name partially matches account_str.

        Args:
            account_str: String used for sub-string matching (case-sensitive).

        Raises: AssertionError if more than one or none of the accounts match
        the account_str.
        """

        matched_accounts = list(
            filter(lambda x: x.count(account_str),
                   [account.name() for account in self.accounts()]))
        if len(matched_accounts) == 0:
            raise AssertionError(f'{account_str} does not match any account '
                                 'in the portfolio')
        if len(matched_accounts) > 1:
            raise AssertionError(f'{account_str} matches more than one '
                                 'account in the portfolio')
        return matched_accounts[0]

    def get_asset_name_by_substr(self, account_str='', asset_str=''):
        """Returns asset names given a sub-string for account and asset.

        Args:
            account_str: String used for sub-string matching with account
            names (case-sensitive).
            asset_str: String used for sub-string matching with asset
            names or asset short names (case-sensitive).

        Returns: A tuple account_name, asset_name where account name
        partially matches account_str and asset name partially matches
        asset_str or asset short name == asset_str

        Raises: AssertionError if none or more than one asset matches the
        sub-strings."""
        matched_assets = list(filter(
            lambda x: x[0].count(account_str) and (
                x[1].count(asset_str) or x[2] == asset_str),
            [(account.name(), asset.name(), asset.short_name())
             for account in self.accounts()
             for asset in account.assets()]))
        if len(matched_assets) < 1:
            raise AssertionError(
                'Provided asset and account strings match none '
                'of the assets.')
        if len(matched_assets) > 1:
            raise AssertionError(
                'Provided asset and account strings match more '
                'than one assets.')
        return matched_assets[0][0], matched_assets[0][2]

    def what_if(self, account_name, asset_name, delta):
        """Changes value of an asset by delta.

        This method runs a whatif scenario if asset_name in account_name
        is changed by delta. The account cash balance is changed with
        -delta to balance the transaction.

        Args:
            account_name: String representing the account name.
            asset_name: String representing the asset name.
            delta: A positive or negative delta for money.
        """
        account = self.get_account(account_name)
        asset = account.get_asset(asset_name)
        asset.what_if(delta)
        # We take the money out of account.
        account.add_cash(-delta)

    def what_if_add_cash(self, account_name, cash_delta):
        """Changes available cash balance of an account by delta."""
        self.get_account(account_name).add_cash(cash_delta)

    def get_what_ifs(self):
        """Returns all what_ifs set by previous methods.

        Returns: A tuple of two table.Table representing
        account whatifs and asset what ifs respectively. The first
        table has columns Account and Cash; and the second table has
        columns Account, Asset and Delta (representing money).
        """
        account_whatifs = Table(
            2,
            headers=['Account', 'Cash'],
            coltypes=['str', 'delta_dollars'])
        asset_whatifs = Table(
            3,
            headers=['Account', 'Asset', 'Delta'],
            coltypes=['str', 'str', 'delta_dollars'])

        for account in self.accounts():
            if account.available_cash() != 0.0:
                account_whatifs.add_row(
                    [account.name(), account.available_cash()])
            for asset in account.assets():
                delta = asset.get_what_if()
                if delta != 0.0:
                    asset_whatifs.add_row(
                        [account.name(), asset.name(), delta])

        return account_whatifs, asset_whatifs

    def reset_what_ifs(self):
        """Reset all previously set what ifs in this portfolio."""
        for account in self.accounts():
            account.add_cash(-account.available_cash())
            for asset in account.assets():
                asset.what_if(-asset.get_what_if())

    def total_value(self):
        """Returns total value of the portfolio."""
        total = 0.0
        for account in self.accounts():
            total += account.available_cash()
            for asset in account.assets():
                total += asset.adjusted_value()
        return total

    # TODO: Renmame this to list_assets for consistency.
    def assets(self, short_name=False, quantity=False):
        """Returns all the assets.

        Args:
            short_name: If set, returns the short name of the asset as well.
            quantity: If set, returns the shares/quantity of the asset for
            the assets that support it.

        Returns: A table.Table object representing all the assets.
        The columns correspond to Account name, short name (optional),
        quantity (optional), asset name and value.
        """
        table = Table(
            3 + short_name + quantity,
            headers=(['Account']
                     + (['Name'] if short_name else [])
                     + (['Quantity'] if quantity else [])
                     + ['Asset', 'Value']),
            coltypes=(['str']
                      + (['str'] if short_name else [])
                      + (['float'] if quantity else [])
                      + ['str', 'dollars']))
        for account in self.accounts():
            for asset in account.assets():
                row = ([account.name()]
                       + ([f'{asset.short_name()}'] if short_name else [])
                       + [asset.name(), asset.adjusted_value()])
                if quantity:
                    row.insert(1 + short_name, asset.shares()
                               if hasattr(asset, 'shares') else None)
                table.add_row(row)
        return table

    def list_lots(self):
        """Returns all the tax lots in the portfolio.

        Returns: A table.Table object representing tax lots
        for assets that support it. The columns of the returned table are
        short name (of asset), date of lot, cost basis of lot, gain (+ve or
        -ve) and percentage gain.
        """
        table = Table(
            5,
            headers=['Short Name', 'Date', 'Cost', 'Gain', 'Gain%'],
            coltypes=['str', 'str', 'dollars', 'delta_dollars', 'percentage'])
        for account in self.accounts():
            for asset in account.assets():
                if hasattr(asset, 'list_lots'):
                    lots = asset.list_lots()
                    assert (
                        lots.headers()
                        == ['Date', 'Quantity', 'Cost', 'Gain', 'Gain%'])
                    for lot in lots.list():
                        table.add_row([asset.short_name()] + lot[:1] + lot[2:])
        return table

    def asset_location(self):
        """Returns asset location of this portfolio.

        Returns: A table.Table object representing the asset location.
        The columns of the table correspond to Asset class name, account type,
        percentage allocation to account type and total monetary value.
        """
        class2type = {}  # Mapping of asset class -> account type -> money
        for account in self.accounts():
            for asset in account.assets():
                for name, ratio in asset.class2ratio.items():
                    if name not in class2type:
                        class2type[name] = {}
                    class2type[name][account.account_type] = (
                        class2type[name].get(account.account_type, 0)
                        + ratio * asset.adjusted_value())

        table = Table(
            4,
            headers=['Asset Class', 'Account Type', 'Percentage', 'Value'],
            coltypes=['str', 'str', 'percentage', 'dollars'])

        for asset_class, type2value in class2type.items():
            first = True
            total = sum(type2value.values())
            if abs(total) < 1e-6:
                continue
            for account_type, value in sorted(
                    type2value.items(),
                    key=lambda x: x[1],
                    reverse=True):
                table.add_row([asset_class if first else '', account_type,
                               value / total, value])
                first = False
        return table

    def _get_asset_class_to_value(self):
        """Returns asset class name -> money allocated to it."""
        asset_class_to_value = {}

        for account in self.accounts():
            for asset in account.assets():
                for name, ratio in asset.class2ratio.items():
                    asset_class_to_value[name] = asset_class_to_value.get(
                        name, 0) + ratio * asset.adjusted_value()

        return asset_class_to_value

    def asset_allocation_tree(self, levels=-1):
        """Returns asset allocation in long vertical tree format.

        Args:
            levels: The max depth of asset allocation tree (-1 = all levels).

        Returns: table.Table object representing the asset allocation. The
        table has multiple sections separated by empty row, each corresponding
        to an asset class and it's children asset class. The first row in a
        section is the asset class name followed by ":". The remaining rows
        correspond to the direct children of the asset class. For these
        rows the columns correspond to asset class name, actual percentage of
        money allocated to it, the desired percentage of money allocated, and
        the actual money allocated to it.
        """
        table = Table(4,
                      headers=['Class', 'Actual%', 'Desired%', 'Value'],
                      coltypes=['str', 'percentage', 'percentage', 'dollars'])
        first_row = True
        for alloc in self.asset_classes.return_allocation(
                self._get_asset_class_to_value(), levels):
            if not alloc.children:
                continue

            if not first_row:
                table.add_row([' '])
            first_row = False
            table.add_row([f'{alloc.name}:'])
            for child in alloc.children:
                table.add_row([child.name,
                              child.actual_allocation,
                              child.desired_allocation,
                              child.value])
        return table

    def asset_allocation(self, asset_class_list):
        """Returns asset allocation across the asset classes provided.

        This method returns the asset allocation, but only across the asset
        classes provided via asset_class_list. This requires that the asset
        class list provided represents a "cut" of the asset class tree. What
        it means is that the asset list should cover the asset tree completely.
        None of the asset classes should be a descendent or ancestor of another
        and the sum of actual allocation across the asset class list must add
        up to 100%.

        Args:
            asset_class_list: A list of strings representing the asset classes
            across which the asset allocation should be returned.

        Returns: A table.Table object representing the asset allocation. The
        columns represent asset class name, actual percentage allocation to
        the asset class, the desired percentage allocation of money to the
        asset class, the absolute money allocated to it and difference from
        the desired amount of money allocated.

        Raises: AssertionError is the asset_class_list is not a proper
        "cut" of the asset class tree.
        """
        flat_asset_class = AssetClass('Root')
        for asset_class in asset_class_list:
            found = self.asset_classes.find_asset_class(asset_class)
            assert found, f'Could not find {asset_class}'
            flat_asset_class.add_subclass(found[1], found[0])

        try:
            flat_asset_class.validate()
        except AssertionError:
            raise AssertionError(
                'AssetAllocation called with overlapping '
                'Asset Classes or Asset Classes which does not cover the '
                'full tree.') from None

        alloc = flat_asset_class.return_allocation(
            self._get_asset_class_to_value(), 0)[0]
        table = Table(
            5,
            headers=['Class', 'Actual%', 'Desired%', 'Value', 'Difference'],
            coltypes=['str', 'percentage', 'percentage', 'dollars',
                      'delta_dollars'])
        for child in alloc.children:
            table.add_row([child.name, child.actual_allocation,
                          child.desired_allocation, child.value,
                          child.value_difference])
        return table

    def asset_allocation_compact(self):
        """Returns asset allocation in long horizontal format.

        This method is similar to asset_allocaton_tree method in terms of
        informaiton returned, but this method returns the asset allocation
        in horizontal format.

        Returns: A table.Table object representing the portfolio's asset
        allocation. Each row corresponds to a path in the asset allocation
        tree. The columns are grouped into groups of three for each node
        in the tree (except the final two columns). The three columns for
        each node correspond to the asset class name, actual percentage
        of assets mapped to the asset class, and the desired percentage
        of assets. The final two columns are corresponding to the leaf
        nodes of asset allocation: They contain the actual value of assets
        mapped to the leaf node, and the difference from the desired
        value of assets mapped to the leaf node, respectively.
        """
        def find_index(class_name, ret_list):
            # We assume ret_list has atleast one entry.
            names_list = [row[-3] for row in ret_list]
            return names_list.index(class_name)

        ret_list = []
        for alloc in self.asset_classes.return_allocation(
                self._get_asset_class_to_value()):
            if not alloc.children:
                continue

            if not ret_list:
                for child in alloc.children:
                    ret_list.append([child.name,
                                     child.actual_allocation,
                                     child.desired_allocation])
            else:  # Parent is already in ret_list
                index = find_index(alloc.name, ret_list)
                # We know that here is atleast one child.
                for i in range(len(alloc.children) - 1):
                    # Make room for rest of the children by inserting empty
                    # extra rows of the same size as the parent's row
                    ret_list.insert(index + 1, [None] * len(ret_list[index]))
                for i in range(len(alloc.children)):
                    ret_list[index + i].extend(
                        [alloc.children[i].name,
                         alloc.children[i].actual_allocation,
                         alloc.children[i].desired_allocation])

        # Return early if there is no AA.
        if not ret_list:
            return Table(0)

        # By this time ret_list has all the AA tree.
        # Add Leaf node AA at the end.
        cols = max(map(len, ret_list))
        leaf_aa = self.asset_allocation(self.asset_classes.leaves())
        for leaf_row in leaf_aa.list():  # Format: Class, A%, D%, Value, Diff
            ret_list_row = ret_list[find_index(leaf_row[0], ret_list)]
            ret_list_row.extend([None] * (cols - len(ret_list_row)))
            ret_list_row.extend(leaf_row[1:])

        # All done, now build the table.
        t = Table(cols + 4,
                  headers=['Class', 'A%', 'D%'] * int(cols / 3)
                  + leaf_aa.headers()[1:],
                  coltypes=['str', 'percentage', 'percentage'] * int(cols / 3)
                  + ['percentage', 'percentage', 'dollars', 'delta_dollars'])
        t.set_rows(ret_list)
        return t



"""Command line interface for Lakshmi.

This is meant to be used as an application and would not work well if used
as library (by design it keeps a lot of global state and is not safe to be
called multiple times from the same program). If there is ever need to use
it as a library, this code requires major refactoring to clean it up."""

from pathlib import Path

import click
import yaml

import lakshmi.analyze
import lakshmi.assets
import lakshmi.cache
from lakshmi import Portfolio
from lakshmi.table import Table


class LakContext:
    """Context class with utilities to help the script keep state and
    share it."""
    DEFAULT_PORTFOLIO = '~/portfolio.yaml'

    def _return_config(self):
        """Internal function to read and return .lakrc file."""
        lakrcfile = Path(self.lakrc).expanduser()

        if not lakrcfile.exists():
            return {}
        config = yaml.load(lakrcfile.read_text(), Loader=yaml.SafeLoader)
        return config

    def __init__(self, lakrc='~/.lakrc'):
        self.lakrc = lakrc
        # Used in self.optional_separator()
        self.continued = False
        # Used in self.warn_for_what_ifs()
        self.warned = False
        # List of hypothetical whatifs. Used in self.get_what_ifs() and
        # self.warn_for_what_ifs()
        self.whatifs = None
        # The loaded portfolio.
        self.portfolio = None
        self.tablefmt = None

        config = self._return_config()

        portfolio_filename = config.pop(
            'portfolio', LakContext.DEFAULT_PORTFOLIO)
        self.portfolio_filename = str(Path(portfolio_filename).expanduser())

        # Setup cache directory.
        if 'cache' in config:
            cache_dir = config.pop('cache')
            if cache_dir is None:
                # Disable caching.
                lakshmi.cache.set_cache_dir(None)
            else:
                lakshmi.cache.set_cache_dir(Path(cache_dir).expanduser())

        if len(config):
            raise click.ClickException('Extra entries found in config file: '
                                       f'{list(config.keys())}')

    def optional_separator(self):
        """Prints a newline between multiple commands. Used to add a newline
        between cmd1, cmd2, cmd3, etc. in 'lak list cmd1 cmd2 cmd3'."""
        # Don't print separator if this is the first time we are called.
        if self.continued:
            click.echo()
        # Set it up so that separator is printed for the next command.
        self.continued = True

    def get_what_ifs(self):
        """Load and return a list of hypothetical whatifs in the portfolio."""
        if not self.whatifs:
            self.whatifs = self.get_portfolio().get_what_ifs()
        return self.whatifs

    def warn_for_what_ifs(self):
        """Prints a warning if whatifs are set."""
        # Make sure we don't print warning multiple times if commands are
        # chained.
        if self.warned:
            return
        self.get_what_ifs()
        if self.whatifs[0].list() or self.whatifs[1].list():
            click.secho('Warning: Hypothetical what ifs are set.\n', fg='red')
        self.warned = True

    def get_portfolio(self):
        """Loads and returns the portfolio from self.portfolio_filename."""
        if not self.portfolio:
            # Check if portfolio file doesn't exist and print helpful error
            # message.
            portfolio_file = Path(self.portfolio_filename).expanduser()
            if not portfolio_file.exists():
                raise click.ClickException(
                    f'Portfolio file {portfolio_file} does not exist. Please '
                    'use "lak init" to create a new portfolio.')

            self.portfolio = Portfolio.load(self.portfolio_filename)
        return self.portfolio

    def save_portfolio(self):
        """Save self.portfolio back to file."""
        self.portfolio.save(self.portfolio_filename)


# Global variable to save and pass context between click commands.
#
# I tried using click's builtin context, but it was too troublesome
# and didn't exactly give us the functionality that I wanted
# -- sarvjeets
lakctx = None


class Spinner:
    """Prints a progress bar on the screen for cache misses."""
    SPINNER = ('▰▱▱▱▱▱▱',
               '▰▰▱▱▱▱▱',
               '▰▰▰▱▱▱▱',
               '▰▰▰▰▱▱▱',
               '▰▰▰▰▰▱▱',
               '▰▰▰▰▰▰▱',
               '▰▰▰▰▰▰▰')

    def __init__(self):
        self._index = 0  # Index within _SPINNER
        self._spinning = False  # Set to true once we start 'spinning'
        self._isatty = True  # Are we connected to tty?
        try:
            self._isatty = click.get_binary_stream('stdout').isatty()
        except Exception:
            self._isatty = False

    def __enter__(self):
        if not self._isatty:  # Not on terminal, do nothing.
            return

        def spin():
            if self._spinning:
                click.echo('\b' * len(Spinner.SPINNER[0]), nl=False)
            else:
                self._spinning = True

            click.echo(Spinner.SPINNER[self._index], nl=False)
            self._index = (self._index + 1) % len(Spinner.SPINNER)

        # Setup so that cache calls spin
        lakshmi.cache.set_cache_miss_func(spin)
        return spin

    def __exit__(self, *args):
        lakshmi.cache.set_cache_miss_func(None)  # Disable spinner
        if self._spinning:
            click.echo('\b' * len(Spinner.SPINNER[0]), nl=False)


@click.group()
@click.version_option()
@click.option('--refresh', '-r', is_flag=True,
              help='Re-fetch all data instead of using previously cached '
              'data. For large portfolios, this would be extremely slow.')
def lak(refresh):
    """lak is a simple command line tool inspired by Bogleheads philosophy.
    Detailed user guide is available at:
    https://sarvjeets.github.io/lakshmi/docs/lak.html"""
    lakshmi.cache.set_force_refresh(refresh)
    global lakctx
    if not lakctx:
        # Setup a new context object for child commands.
        lakctx = LakContext()


@lak.group(chain=True)
@click.option('--format', '-f',
              type=click.Choice(
                  ['plain', 'simple', 'github', 'grid', 'fancy_grid',
                   'pipe', 'orgtbl', 'rst', 'mediawiki', 'html', 'latex',
                   'latex_raw', 'latex_booktabs', 'latex_longtable', 'tsv'],
                  case_sensitive=False),
              default='simple',
              help='Set output table format. For more information on table '
              'formats, please see "Table format" section on: '
              'https://pypi.org/project/tabulate/')
def list(format):
    """Command to list various parts of the portfolio."""
    global lakctx
    lakctx.tablefmt = format


@list.command()
def total():
    """Prints the total value of the portfolio."""
    global lakctx
    lakctx.optional_separator()
    lakctx.warn_for_what_ifs()
    portfolio = lakctx.get_portfolio()

    with Spinner():
        output = Table(2, coltypes=['str', 'dollars']).add_row(
            ['Total Assets', portfolio.total_value()]).string(lakctx.tablefmt)
    click.echo(output)


@list.command()
def al():
    """Prints the Asset Location of the portfolio. For more information,
    please see
    https://www.bogleheads.org/wiki/Tax-efficient_fund_placement"""
    global lakctx
    lakctx.optional_separator()
    lakctx.warn_for_what_ifs()
    portfolio = lakctx.get_portfolio()
    with Spinner():
        output = portfolio.asset_location().string(lakctx.tablefmt)
    click.echo(output)


@list.command()
@click.option('--compact/--no-compact', default=True, show_default=True,
              help='Print the Asset allocation tree in a vertically '
              'compact format')
@click.option('--asset-class', '-c', type=str,
              help='If provided, only print asset allocation for these asset '
              'classes. This is a comma separated list of asset classes (not '
              'necessarily leaf asset classes) and the allocation across '
              'these asset classes should sum to 100%.')
def aa(compact, asset_class):
    """Prints the Asset Allocation of the portfolio. For more information,
    please see https://www.bogleheads.org/wiki/Asset_allocation"""
    global lakctx
    lakctx.optional_separator()
    lakctx.warn_for_what_ifs()

    portfolio = lakctx.get_portfolio()
    with Spinner():
        if asset_class:
            if not compact:
                raise click.UsageError(
                    '--no-compact is only supported when --asset-class '
                    'is not specified.')
            classes_list = [c.strip() for c in asset_class.split(',')]
            output = portfolio.asset_allocation(classes_list).string(
                lakctx.tablefmt)
        else:
            if compact:
                output = portfolio.asset_allocation_compact().string(
                    lakctx.tablefmt)
            else:
                output = portfolio.asset_allocation_tree().string(
                    lakctx.tablefmt)
    click.echo(output)


@list.command()
@click.option('--short-name', '-s', is_flag=True,
              help='Print the short name of the assets as well (e.g. Ticker '
              'for assets that have it).')
@click.option('--quantity', '-q', is_flag=True,
              help='Print the quantity of the asset (e.g. quantity/shares '
              'for assets that have it).')
def assets(short_name, quantity):
    """Prints all assets in the portfolio and their current values."""
    global lakctx
    lakctx.optional_separator()
    lakctx.warn_for_what_ifs()
    portfolio = lakctx.get_portfolio()
    with Spinner():
        output = portfolio.assets(
            short_name=short_name, quantity=quantity).string(lakctx.tablefmt)
    click.echo(output)


@list.command()
def whatifs():
    """Prints hypothetical what ifs for assets and accounts."""
    global lakctx
    account_whatifs, asset_whatifs = lakctx.get_what_ifs()
    if account_whatifs.list():
        lakctx.optional_separator()
        click.echo(account_whatifs.string(lakctx.tablefmt))
    if asset_whatifs.list():
        lakctx.optional_separator()
        click.echo(asset_whatifs.string(lakctx.tablefmt))
    lakctx.warned = True  # Don't warn about whatifs if command is chained.


@list.command()
def lots():
    """Prints tax lot information for all the assets."""
    global lakctx
    lakctx.optional_separator()
    with Spinner():
        output = lakctx.get_portfolio().list_lots().string(lakctx.tablefmt)
    if output:
        click.echo(output)


@lak.group(chain=True,
           invoke_without_command=True)
@click.option('--reset', '-r', is_flag=True,
              help='Reset all hypothetical whatif amounts.')
def whatif(reset):
    """Run hypothetical what if scenarios by modifying the total value of
    an account or asset. This is useful to see how the asset allocation
    or location will change if you make these changes. Once you are done
    playing around with the hypothetical changes, you can reset them all
    by using the --reset flag."""
    if reset:
        global lakctx
        lakctx.get_portfolio().reset_what_ifs()
        lakctx.save_portfolio()


# ignore_unknown_options is there to make sure -100 is parsed as delta
# and click doesn't think it is an option.
@whatif.command(context_settings={"ignore_unknown_options": True})
@click.option('--account', '-t', type=str, metavar='substr', required=True,
              help='Change the value of account that matches this substring')
@click.argument('delta', type=float, required=True)
def account(account, delta):
    """Run hypothetical what if scenario on an account.
    This command adds DELTA to the value of account specified."""
    global lakctx
    portfolio = lakctx.get_portfolio()
    account_name = portfolio.get_account_name_by_substr(account)
    portfolio.what_if_add_cash(account_name, delta)
    lakctx.save_portfolio()


# ignore_unknown_options is there to make sure -100 is parsed as delta
# and click doesn't think it is an option.
@whatif.command(context_settings={"ignore_unknown_options": True})
@click.option('--asset', '-a', type=str, metavar='substr', required=True,
              help='Change the value of asset that matches this substring')
@click.option('--account', '-t', type=str, metavar='substr',
              help='If the asset name is not unique across the portfolio, '
              'an optional substring to specify the account to which the '
              'asset belongs.')
@click.argument('delta', type=float, required=True)
def asset(asset, account, delta):
    """Run hypothetical what if scenario on an asset.
    This command adds DELTA to the value of the asset specified."""
    global lakctx
    portfolio = lakctx.get_portfolio()
    account_name, asset_name = portfolio.get_asset_name_by_substr(
        account if account is not None else '', asset)
    portfolio.what_if(account_name, asset_name, delta)
    lakctx.save_portfolio()


@lak.group(chain=True)
def info():
    """Print detailed information about an asset or account."""
    pass


@info.command()
@click.option('--account', '-t', type=str, metavar='substr', required=True,
              help='Print info about the account that matches this substring')
def account(account):
    """Print details of an account."""
    global lakctx
    lakctx.optional_separator()

    portfolio = lakctx.get_portfolio()
    account_name = portfolio.get_account_name_by_substr(account)
    click.echo(portfolio.get_account(account_name).string())


@info.command()
@click.option('--asset', '-a', type=str, metavar='substr', required=True,
              help='Print info about the asset that matches this substring')
@click.option('--account', '-t', type=str, metavar='substr',
              help='If the asset name is not unique across the portfolio, '
              'an optional substring to specify the account to which the '
              'asset belongs.')
def asset(asset, account):
    """Print details of an asset."""
    global lakctx
    lakctx.optional_separator()

    portfolio = lakctx.get_portfolio()
    account_name, asset_name = portfolio.get_asset_name_by_substr(
        account if account is not None else '', asset)
    click.echo(portfolio.get_account(
        account_name).get_asset(asset_name).string())


@lak.group()
def edit():
    """Edit parts of the portfolio."""
    pass


_HELP_MSG_PREFIX = ('\n\n# # Lines starting with "#" are ignored and an '
                    'empty message aborts this command.\n\n')


def edit_and_parse(edit_dict, parse_fn, filename):
    """Helper funtion to edit parts of portfolio.

    This function is used to add/edit parts of the portfolio. It converts
    a dictionary representation of the object into a YAML file that the
    user can edit. The file is then converted back to a dictionary, which
    is parsed via the parse_fn to convert it into an object.

    Arguments:
        edit_dict: Dictionary representing an object that is being
        editted. This is usually the output of to_dict() of an object.
        parse_fn: The function used to parse the resulting dict generated
        by parsing the yaml file.
        filename: The file containing a template YAML file for the object
        being editted (inside lakshmi/data directory). If edit_dict is
        empty, this file is presented to the user so that the user can make
        changes inline. If edit_dict is set, this file is added as a
        comment at the end of the edit_dict representation as a helpful
        guide for the user.
    """
    # Change filename to absolute path.
    filepath = Path(__file__).resolve().parent / 'data' / filename
    if edit_dict:
        help_msg = _HELP_MSG_PREFIX + '# ' + filepath.read_text().replace(
            '\n', '\n# ')
        edit_str = yaml.dump(edit_dict, sort_keys=False) + help_msg
    else:
        edit_str = filepath.read_text()

    while True:
        edit_str = click.edit(edit_str)
        if not edit_str:  # No changes or empty string.
            raise click.Abort()
        try:
            parse_str = (edit_str.split(help_msg, 1)[0].rstrip('\n')
                         if edit_dict else edit_str)
            return parse_fn(yaml.load(parse_str, Loader=yaml.SafeLoader))
        except Exception as e:
            click.echo('Error parsing file: ' + repr(e))
            if not click.confirm('Do you want to edit again?'):
                raise click.Abort()


@lak.command()
def init():
    """Initializes a new portfolio by adding asset classes. This command can
    be used to create an empty portfolio file if one doesn't exist."""
    global lakctx
    if Path(lakctx.portfolio_filename).exists():
        raise click.ClickException(
            f'Portfolio file already exists: {lakctx.portfolio_filename}')

    asset_class = edit_and_parse(
        None,
        lambda x: lakshmi.AssetClass.from_dict(x).validate(),
        'AssetClass.yaml')
    lakctx.portfolio = Portfolio(asset_class)
    lakctx.save_portfolio()


@edit.command()
def assetclass():
    """Edit the Asset classes and the desired asset allocation."""
    global lakctx
    portfolio = lakctx.get_portfolio()

    asset_classes = edit_and_parse(
        portfolio.asset_classes.to_dict(),
        lambda x: lakshmi.AssetClass.from_dict(x).validate(),
        'AssetClass.yaml')
    portfolio.asset_classes = asset_classes
    lakctx.save_portfolio()


@edit.command()
@click.option('--account', '-t', type=str, metavar='substr', required=True,
              help='Edit the account that matches this substring')
def account(account):
    """Edit an account in the portfolio."""
    global lakctx
    portfolio = lakctx.get_portfolio()

    account_name = portfolio.get_account_name_by_substr(account)
    account_obj = portfolio.get_account(account_name)
    # Save assets and restore them after the user is done editing.
    assets = account_obj.assets()
    account_obj.set_assets([])

    account_obj = edit_and_parse(
        account_obj.to_dict(),
        lakshmi.Account.from_dict,
        'Account.yaml')
    account_obj.set_assets(assets)

    if account_obj.name() != account_name:
        portfolio.remove_account(account_name)
    portfolio.add_account(account_obj, replace=True)
    lakctx.save_portfolio()


@edit.command()
@click.option('--asset', '-a', type=str, metavar='substr', required=True,
              help='Edit the asset whose name or short name matches this '
              'substring.')
@click.option('--account', '-t', type=str, metavar='substr',
              help='If the asset name is not unique across the portfolio, '
              'an optional substring to specify the account to which the '
              'asset belongs.')
def asset(asset, account):
    """Edit an asset in the portfolio."""
    global lakctx
    portfolio = lakctx.get_portfolio()

    account_name, asset_name = portfolio.get_asset_name_by_substr(
        account if account is not None else '', asset)
    account_obj = portfolio.get_account(account_name)
    asset_obj = account_obj.get_asset(asset_name)

    asset_obj = edit_and_parse(asset_obj.to_dict(),
                               asset_obj.from_dict,
                               asset_obj.__class__.__name__ + '.yaml')

    if asset_obj.short_name() != asset_name:
        account_obj.remove_asset(asset_name)
    account_obj.add_asset(asset_obj, replace=True)

    lakctx.save_portfolio()


@lak.group()
def add():
    """Add new accounts or assets to the portfolio."""
    pass


@add.command()
def account():
    """Add a new account to the portfolio."""
    global lakctx
    portfolio = lakctx.get_portfolio()

    account = edit_and_parse(None, lakshmi.Account.from_dict, 'Account.yaml')
    portfolio.add_account(account)
    lakctx.save_portfolio()


@add.command()
@click.option('--asset-type', '-p', required=True,
              type=click.Choice([c.__name__ for c in lakshmi.assets.CLASSES],
                                case_sensitive=False),
              help='Add this type of asset.')
@click.option('--account', '-t', type=str, metavar='substr', required=True,
              help='Add asset to this account (a substring that matches the '
              'account name).')
def asset(asset_type, account):
    """Edit assets in the portfolio."""
    global lakctx
    portfolio = lakctx.get_portfolio()

    account_name = portfolio.get_account_name_by_substr(account)
    account_obj = portfolio.get_account(account_name)
    asset_cls = [
        c for c in lakshmi.assets.CLASSES if c.__name__ == asset_type][0]

    asset_obj = edit_and_parse(None, asset_cls.from_dict, asset_type + '.yaml')
    account_obj.add_asset(asset_obj)

    lakctx.save_portfolio()


@lak.group()
def delete():
    """Delete an account or asset."""
    pass


@delete.command()
@click.option('--account', '-t', type=str, metavar='substr', required=True,
              help='Delete the account that matches this substring')
@click.confirmation_option(prompt='This operation is not reversable. '
                           'Are you sure?')
def account(account):
    """Delete an account from the portfolio."""
    global lakctx
    portfolio = lakctx.get_portfolio()
    account_name = portfolio.get_account_name_by_substr(account)
    portfolio.remove_account(account_name)
    lakctx.save_portfolio()


@delete.command()
@click.option('--asset', '-a', type=str, metavar='substr', required=True,
              help='Delete the asset whose name or short name matches this '
              'substring.')
@click.option('--account', '-t', type=str, metavar='substr',
              help='If the asset name is not unique across the portfolio, '
              'optionally a substring to specify the account name '
              'from which the asset should be deleted.')
@click.confirmation_option(prompt='This operation is not reversable. '
                           'Are you sure?')
def asset(asset, account):
    """Delete an asset from the portfolio."""
    global lakctx
    portfolio = lakctx.get_portfolio()

    account_name, asset_name = portfolio.get_asset_name_by_substr(
        account if account is not None else '', asset)
    account_obj = portfolio.get_account(account_name)
    account_obj.remove_asset(asset_name)
    lakctx.save_portfolio()


@lak.group(chain=True)
def analyze():
    """Analyze the portfolio. """
    pass


@analyze.command()
@click.option('--max-percentage', '-p', type=float, default=10,
              show_default=True,
              help='The max percentage loss for each lot before TLHing.')
@click.option('--max-dollars', '-d', type=int,
              help='The max absolute loss for an asset (across all tax lots) '
              'before TLHing.')
def tlh(max_percentage, max_dollars):
    """Shows which tax lots can be Tax-loss harvested (TLH)."""
    global lakctx
    lakctx.optional_separator()
    with Spinner():
        table = lakshmi.analyze.TLH(max_percentage / 100, max_dollars).analyze(
            lakctx.get_portfolio())
    if not table.list():
        click.echo('No tax lots to harvest.')
    else:
        click.echo(table.string())


@analyze.command()
@click.option('--max-abs-percentage', '-a', type=float, default=5,
              show_default=True,
              help='Max absolute difference before rebalancing.')
@click.option('--max-relative-percentage', '-r', type=float, default=25,
              help='The max relative difference before rebalancing.')
def rebalance(max_abs_percentage, max_relative_percentage):
    """Shows if assets needs to be rebalanced based on a band
    based rebalancing scheme. For more information, please refer to
    https://www.whitecoatinvestor.com/rebalancing-the-525-rule/."""
    global lakctx
    lakctx.optional_separator()
    with Spinner():
        table = lakshmi.analyze.BandRebalance(
            max_abs_percentage / 100, max_relative_percentage / 100).analyze(
            lakctx.get_portfolio())
    if not table.list():
        click.echo('Portfolio Asset allocation within bounds.')
    else:
        click.echo(table.string())


if __name__ == '__main__':
    lak()



"""
This class is used to cache return value of functions on disk for a specified
number of days. This is used by lakshmi.assets module to cache name/ asset
value (i.e the slow functions). For examples on how to use this class, please
see the tests (tests/test_cache.py file).

Currently, this module can only be used on functions which are class members
and the function itself must take no arguments. These restrictions can be
easily relaxed, but so far that all usecases don't need anything more than what
is currently implemented.

In addition to caching values, this class also allows one to optionally call
a user-specified function on cache-misses (currently used to show a progress
bar to the user via the lak CLI).
"""

import functools
import pickle
from abc import ABC, abstractmethod
from datetime import datetime
from hashlib import md5
from pathlib import Path

# Inspired by https://pypi.org/project/cache-to-disk/. I tried using other
# options such as requests-cache, but it was too slow compared to the solution
# implemented here.
# TODO(sarvjeets): It would be good to get rid of this one-off solution and
# switch to something more standard.


class Cacheable(ABC):
    """Interface that declares that a particular class's method return
    values could be cached. The methods should not take a parameter,
    and cache_key() + method name should uniquely imply the return
    value of that class."""
    @abstractmethod
    def cache_key(self):
        """Unique string value used as key for caching."""
        pass


def get_file_age(file):
    """Returns the age of file.

    Args:
        file: A PosixPath object representing a file.

    Returns: An int represeting the age in days.
    """
    return (datetime.today()
            - datetime.fromtimestamp(file.stat().st_mtime)).days


# Constants
# Default cache directory if none is specified.
_DEFAULT_DIR = Path.home() / '.lakshmicache'
_CACHE_STR = 'cache_dir'
_FORCE_STR = 'force_refresh'
_MISS_FUNC_STR = 'miss_func'

# Dict (string -> object) to keep cache context.
# Description of keys to what is stored:
# _CACHE_STR:
# The pathlib.Path object specifying cache directory. If set to None,
# caching is disabled. Default: _DEFAULT_DIR
# _FORCE_STR:
# If set to True, new values are generated even if a cached one is
# available. Default: False
# _MISS_FUNC_STR:
# If set, this function is called for every cache miss.
_ctx = {_FORCE_STR: False}


def set_force_refresh(v):
    """Sets whether cached values should be refreshed.

    Args:
        v: Boolean representing if cached values should be re-generated.
    """
    global _ctx
    _ctx[_FORCE_STR] = v


def set_cache_miss_func(f):
    """Sets the function to call for cache-misses.

    Args:
        f: The function to call whenever a cache-miss happens (i.e. whenever
        the underlying function is called instead of using a cached value).
    """
    global _ctx
    if f:
        _ctx[_MISS_FUNC_STR] = f
    else:
        # Clear out previously set function, if any.
        _ctx.pop(_MISS_FUNC_STR, None)


def set_cache_dir(cache_dir):
    """Sets the cache directory.

    If the cache directory is not specified, default ~/.lakshmicache
    is used.

    Args:
        cache_dir: The pathlib.Path object specifying cache directory.
        If set to None, caching is disabled.
    """
    global _ctx
    _ctx[_CACHE_STR] = cache_dir
    if cache_dir is None:
        return
    cache_dir.mkdir(exist_ok=True)  # Create cache dir if one doesn't exist.
    # Delete old files whose cache values are invalid already.
    for file in cache_dir.glob('*_*.lkc'):
        days = int(file.name.split('_')[0])
        if get_file_age(file) >= days:
            file.unlink()


def call_func(class_obj, func):
    """Helper function to return value of class_obj.func().

    In addition to calling function, this helper also calls the
    cache_miss function if one is set in the context.

    Args:
        class_obj: The object of a particular class implementing Cacheable
        interface.
        func: The function whose return values has to be cached. Assumed
        to take no parameters.

    Returns: The return value of the func.
    """
    global _ctx
    if _MISS_FUNC_STR in _ctx:
        _ctx[_MISS_FUNC_STR]()
    return func(class_obj)


def cache(days):
    """Returns decorator that caches functions return value on disk for
    specified number of days.

    Args:
        days: Number of days for which to cache the return value of the
        function.

    Returns: The decorator.
    """
    def decorator(func):
        @functools.wraps(func)
        def new_func(class_obj):
            global _ctx
            if _CACHE_STR not in _ctx:
                # Cache dir not set. Set to default.
                set_cache_dir(_DEFAULT_DIR)
            cache_dir = _ctx[_CACHE_STR]
            if not cache_dir:
                return call_func(class_obj, func)
            force_refresh = _ctx[_FORCE_STR]

            key = f'{func.__qualname__}_{class_obj.cache_key()}'
            filename = f'{days}_{md5(key.encode("utf8")).hexdigest()}.lkc'
            file = cache_dir / filename

            if (
                not force_refresh
                and file.exists()
                and get_file_age(file) < days
            ):
                return pickle.loads(file.read_bytes())

            value = call_func(class_obj, func)
            file.write_bytes(pickle.dumps(value))
            return value
        return new_func
    return decorator



"""Module to help output and print tables."""

from tabulate import tabulate

import lakshmi.utils as utils


class Table():
    """This class helps format, process and print tabular information."""
    # Mapping from column type to a function that formats the cell
    # entries and converts them to a string.
    # Standard coltypes are:
    # 'str': Column is already in string format.
    # 'dollars': Column is a float which represents a dollar value.
    # 'delta_dollars': Column represents a positive or negative dollar
    # difference.
    # 'percentage': A float representing a percentage.
    # 'float': Float.
    coltype2func = {
        'str': lambda x: x,
        'dollars': lambda x: utils.format_money(x),
        'delta_dollars': lambda x: utils.format_money_delta(x),
        'percentage': lambda x: f'{round(100*x)}%',
        'float': lambda x: str(float(x)),
    }

    # Mapping of column type to how it should be aligned. Most values are
    # self explanatory. 'float' is aligned on the decimal point.
    coltype2align = {
        'str': 'left',
        'dollars': 'right',
        'delta_dollars': 'right',
        'percentage': 'right',
        'float': 'decimal',
    }

    def __init__(self, numcols, headers=(), coltypes=None):
        """
        Args:
            numcols: Number of columns (required)
            headers: Header row (optional)
            coltypes: The type of columns, if not provided the columns are
            assumed to be strings.
        """
        assert numcols >= 0
        self._numcols = numcols

        if headers:
            assert len(headers) == numcols
        self._headers = headers

        if coltypes:
            assert len(coltypes) == numcols
            assert set(coltypes).issubset(
                Table.coltype2func.keys()), 'Bad column type in coltypes'
            self._coltypes = coltypes
        else:
            self._coltypes = ['str'] * self._numcols

        self._rows = []

    def add_row(self, row):
        """Add a new row to the table.

        Args:
            row: A list of column entries representing a row.
        """
        assert len(row) <= self._numcols
        self._rows.append(row)
        return self

    def set_rows(self, rows):
        """Replaces all rows of this table by rows.

        Args:
            rows: A list (rows) of list (columns) of cell entries.
        """
        assert max(map(len, rows)) <= self._numcols
        self._rows = rows

    def headers(self):
        """Returns the header row."""
        return self._headers

    def col_align(self):
        """Returns the column alignment parameters.

        Returns: A list of strings, where each value represents
        the value of coltype2align map. These alignment parameters are
        dependent on the column types specified while constructing this
        object.
        """
        return list(map(lambda x: Table.coltype2align[x], self._coltypes))

    def list(self):
        """Returns the table as a list (row) of lists (raw columns).

        This function doesn't perform any string conversion on the cell values.
        """
        return self._rows

    def str_list(self):
        """Returns the table as a list (row) of list of strings (columns).

        This function converts the raw value of a cell to string based on its
        column type.
        """
        ret_list = []
        for row in self.list():
            ret_row = []
            for col_num in range(len(row)):
                if row[col_num] is None:
                    ret_row.append('')
                else:
                    ret_row.append(Table.coltype2func[self._coltypes[col_num]](
                        row[col_num]))
            ret_list.append(ret_row)
        return ret_list

    def string(self, tablefmt='simple'):
        """Returns the table as a formatted string."""
        str_list = self.str_list()
        if not str_list:
            return ''

        return tabulate(str_list,
                        headers=self.headers(),
                        tablefmt=tablefmt,
                        colalign=self.col_align())



"""This module contains implementation of all the standard asset types.
The top-level interface which every asset type must implement is
lakshmi.assets.Asset. This class also contains helper functions that
operate on an asset type.
"""

import datetime
import re
from abc import ABC, abstractmethod

import requests
import yfinance

import lakshmi.constants
import lakshmi.utils as utils
from lakshmi.cache import Cacheable, cache
from lakshmi.table import Table


def to_dict(asset):
    """Returns a dictionary representation of asset.

    This function is used to convert an 'asset' (aka a class implementing
    lakshmi.assets.Asset interface) into a dictionary. This is mainly used
    to serialize an asset type into a yaml-friendly format. The function
    lakshmi.assets.from_dict is an inverse of this function.

    Args:
        asset: An object of a class implementating lakshmi.assets.Asset
        interface.
    Returns: A dictionary representation of asset.
    """
    return {asset.__class__.__name__: asset.to_dict()}


def from_dict(d):
    """Converts a dictionary representing an asset into an asset object.

    This function is reverse of lakshmi.assets.to_dict function.

    Args:
        d: A dictionary representating an asset type.

    Returns: An object of class implementing lakshmi.assets.Asset interface
    corresponding to d.

    Raises: AssertionError if dict doesn't represent a lakshmi asset type.
    """
    keys = list(d.keys())
    assert len(keys) == 1
    class_name = keys[0]

    for c in CLASSES:
        if c.__name__ == class_name:
            return c.from_dict(d.pop(class_name))

    raise AssertionError(f'Class {class_name} not found.')


class Asset(ABC):
    """Top-level class representing an asset (fund, ETF, cash, etc.).

    Every asset type in lakshmi must inherit from this class.
    """

    def __init__(self, class2ratio):
        """
        Args:
          class2ratio: Dict of class_name -> ratio, where 0 < ratio <= 1.0

        Raises: AssertionError if ratio is not in (0, 1] or if the sum of
        ratio across all class_name is not equal to 1.
        """
        self._delta = 0
        self.class2ratio = class2ratio

        total = 0
        for ratio in class2ratio.values():
            assert ratio > 0.0 and ratio <= 1.0, (
                f'Bad Class ratio provided to Asset ({ratio})')
            total += ratio

        assert abs(total - 1.0) < 1e-6, (
            'Total allocation to classes must be 100% (actual = '
            f'{total * 100}%)')

    def to_dict(self):
        """Encodes this object into a dictionary.

        Convention: This method for non-abstract (leaf) Asset classes encodes
        all the fields present in the object. This method for abstract Asset
        classes only encodes fields that are not passed in the constructor
        during initialization of the object (encoding those fields is the
        responsibility of the sub-class).

        Returns: A dictionary object representing self.
        """
        if self._delta != 0:
            return {'What if': self._delta}
        return dict()

    def from_dict(self, d):
        """Reverse of to_dict.

        This function initializes the object of this class given a dictionary.
        Convention: This method for non-abstract Asset classes is a factory
        method (static function) that initializes and returns a new object.
        This method for abstract Asset classes decodes non-constructor data (if
        any) and modifies self.

        Args:
            d: A dictionary representing self.

        Returns: An object initialized with fields present in d.
        """
        self.what_if(d.pop('What if', 0))
        return self

    def to_table(self):
        """Returns a table representing this object.

        This function converts self object into a table
        suitable for pretty-printing.

        Returns: lakshmi.table.Table object representing the data in this
        class.
        """
        table = Table(2).add_row(['Name:', f'{self.name()}'])

        asset_mapping_table = Table(2, coltypes=['str', 'percentage'])
        for asset_class, ratio in self.class2ratio.items():
            asset_mapping_table.add_row([f'{asset_class}', ratio])
        table.add_row(['Asset Class Mapping:',
                       f'{asset_mapping_table.string(tablefmt="plain")}'])

        if self._delta:
            table.add_row(['Adjusted Value:',
                           f'{utils.format_money(self.adjusted_value())}'])
            table.add_row(
                ['What if:', f'{utils.format_money_delta(self._delta)}'])
        else:
            table.add_row(
                ['Value:', f'{utils.format_money(self.adjusted_value())}'])

        return table

    def string(self):
        """Returns a string representation of this object."""
        return self.to_table().string(tablefmt='plain')

    def what_if(self, delta):
        """Adds delta (what if) to the adjusted value of this asset.

        This function is provided to all assets and adds delta
        to the adjusted value (without changing number of shares, etc.). This
        is used to check how the asset allocation etc. would change if the
        value of this asset is changed. After verifying the changes, it's
        really easy to undo the changes by a call to what_if(-delta). A
        bit of warning: what_if(100) followed by what_if(100) will add
        200 to the value (i.e. the call doesn't reset any existing deltas
        that are already added to the value.

        Args:
            delta: A float to be added to the total value.
        """
        self._delta += delta
        if abs(self._delta) < 1e-6:
            self._delta = 0

    def get_what_if(self):
        """Returns any what ifs (delta) that are added to the value."""
        return self._delta

    def adjusted_value(self):
        """Returns the adjusted value (after adding any what ifs)."""
        return max(0, self.value() + self.get_what_if())

    @abstractmethod
    def value(self):
        """Returns the value of this asset."""
        pass

    @abstractmethod
    def name(self):
        """Returns the full name of this asset."""
        pass

    @abstractmethod
    def short_name(self):
        """Returns the short name of this asset (ideally < 10 chars)."""
        pass


class ManualAsset(Asset):
    """This is a special catch-all asset type that represents an asset whose
    value is manually specified and is not updated automatically."""
    def __init__(self, name, value, class2ratio):
        """
        Args:
            name: Full name of this asset (also is used as short name).
            value: The current value (in dollars, float) of this asset.
            class2ratio: Dict of class_name -> ratio, where 0 < ratio <= 1.0

        Raises: AssertionError if value is negative.
        """
        assert value >= 0, 'Value of an asset can not be negative.'
        self._name = name
        self._value = value
        super().__init__(class2ratio)

    def to_dict(self):
        """Returns a dict representing this object."""
        d = {'Name': self._name,
             'Value': self._value,
             'Asset Mapping': self.class2ratio}
        d.update(super().to_dict())
        return d

    @classmethod
    def from_dict(cls, d):
        """Returns a new object specified by dictionary d.

        This is reverse of to_dict.
        Args:
            d: A dictionary (usually the output of to_dict).

        Returns: A new ManualAsset object.

        Raises: AssertionError if d cannot be parsed correctly.
        """
        ret_obj = ManualAsset(d.pop('Name'),
                              d.pop('Value', 0),
                              d.pop('Asset Mapping'))
        Asset.from_dict(ret_obj, d)
        assert len(d) == 0, f'Extra attributes found: {list(d.keys())}'
        return ret_obj

    def value(self):
        """Returns value of this asset."""
        return self._value

    def name(self):
        """Returns name of this asset."""
        return self._name

    def short_name(self):
        """Returns short name (same as name) of this asset."""
        return self._name


class TaxLot:
    """Class representing a single tax lot for an Asset."""

    def __init__(self, date, quantity, unit_cost):
        """
        Args:
            date: String representing date (YYYY/MM/DD) on which this lot was
            bought.
            quantity: Number of shares bought on date.
            unit_cost: Price per share.

        Raises: AssertionError if date is not in the right format.
        """
        # Do some sanity check.
        date_pattern = re.compile('\\d{4}/\\d{2}/\\d{2}')
        assert date_pattern.match(
            date), 'Tax lot dates should be in format YYYY/MM/DD'

        self.date = date
        self.quantity = quantity
        self.unit_cost = unit_cost

    def to_dict(self):
        """Converts this object into a dictionary."""
        return {'Date': self.date,
                'Quantity': self.quantity,
                'Unit Cost': self.unit_cost}

    @classmethod
    def from_dict(cls, d):
        """Factory method to return a new object representing a dictionary.

        This is reverse of to_dict. This function returns a newly initialized
        TaxLot object corresponding to d.

        Args:
            d: A dictionary representing TaxLot (usually output of to_dict)

        Returns: An initialied TaxLot object corresponding to d.

        Raises: AssertionError if d can't be parsed properly.
        """
        ret_obj = TaxLot(d.pop('Date'), d.pop('Quantity'), d.pop('Unit Cost'))
        assert len(d) == 0, f'Extra attributes found: {list(d.keys())}'
        return ret_obj


class TradedAsset(Asset):
    """Abstract class representing an asset that is traded on stock market.

    This asset is assumed to have 'shares' and per unit price and can
    optionally have tax lots."""

    def __init__(self, shares, class2ratio):
        """
        Args:
            shares: Number of shares of this asset.
            class2ratio: Dict of class_name -> ratio, where 0 < ratio <= 1.0
        """
        self._shares = shares
        self._tax_lots = None
        super().__init__(class2ratio)

    def to_dict(self):
        """Converts this asset into a dictionary."""
        d = dict()
        if self._tax_lots:
            d.update({'Tax Lots': [lot.to_dict() for lot in self._tax_lots]})
        d.update(super().to_dict())
        return d

    def from_dict(self, d):
        """Initializes self with data provided via dictionary d."""
        super().from_dict(d)
        if 'Tax Lots' not in d:
            return
        tax_lots_list = [TaxLot.from_dict(lot_dict)
                         for lot_dict in d.pop('Tax Lots')]
        self.set_lots(tax_lots_list)
        return self

    def shares(self):
        """Returns the number of shares."""
        return self._shares

    def get_lots(self):
        """Returns the tax lots or None if they are not set.

        Returns: A list of TaxLot or None if not set.
        """
        return self._tax_lots

    def set_lots(self, tax_lots_list):
        """Sets the tax lots.

        Args:
            tax_lots_list: A list of TaxLot representing all the tax_lots.

        Raises: AssertionError if the number of shares in the lots don't
        sum up to the number of shares in this asset.
        """
        sum_lots = sum([t.quantity for t in tax_lots_list])
        assert abs(sum_lots - self._shares) < 1e-6, (
            f'Lots provided should sum up to {self._shares}')
        self._tax_lots = tax_lots_list
        return self

    def list_lots(self):
        """Returns a table of tax lots.

        This function returns a Table of tax lots which can be used to
        pretty-print the tax lot information.

        Returns: lakshmi.table.Table object containing Date, Quantity,
        Cost, Gain and Gain% fields for all the lots.
        """
        table = Table(5,
                      headers=['Date', 'Quantity', 'Cost', 'Gain', 'Gain%'],
                      coltypes=['str', 'float', 'dollars', 'delta_dollars',
                                'percentage'])
        if not self._tax_lots:
            return table

        for lot in self._tax_lots:
            table.add_row(
                [lot.date,
                 lot.quantity,
                 lot.unit_cost * lot.quantity,
                 (self.price() - lot.unit_cost) * lot.quantity,
                 self.price() / lot.unit_cost - 1])
        return table

    def to_table(self):
        """Returns this asset as a lakshmi.table.Table object."""
        table = super().to_table()
        table.add_row(['Price:', f'{utils.format_money(self.price())}'])
        return table

    def string(self):
        """Returns this asset as a string."""
        if not self._tax_lots:
            return super().string()

        return (super().string() + '\n\nTax lots:\n'
                + f'{self.list_lots().string()}')

    def value(self):
        """Returns the current market value of this asset."""
        return self.shares() * self.price()

    # This class inherits abstract methods Name & short_name from Asset.

    @abstractmethod
    def price(self):
        """Returns the current market value of this asset."""
        pass


class NotFoundError(Exception):
    pass


class TickerAsset(TradedAsset, Cacheable):
    """An asset class represented by a Ticker whose price can be pulled."""

    def __init__(self, ticker, shares, class2ratio):
        """
        Args:
            ticker: Ticker of this asset (string),
            shares: Total number of shares.
            class2ratio: Dict of class_name -> ratio, where 0 < ratio <= 1.0
        """
        self._ticker = ticker
        session = requests.Session()
        session.headers['user-agent'] = (
            f'{lakshmi.constants.NAME}/{lakshmi.constants.VERSION}')
        self.yticker = yfinance.Ticker(ticker, session=session)
        super().__init__(shares, class2ratio)

    def to_dict(self):
        """Returns a dict representing this object."""
        d = {'Ticker': self._ticker,
             'Shares': self.shares(),
             'Asset Mapping': self.class2ratio}
        d.update(super().to_dict())
        return d

    @classmethod
    def from_dict(cls, d):
        """Returns a new object specified by dictionary d.

        This is reverse of to_dict.
        Args:
            d: A dictionary (usually the output of to_dict).

        Returns: A new TradedAsset object.

        Raises: AssertionError if d cannot be parsed correctly.
        """
        ret_obj = TickerAsset(
            d.pop('Ticker'),
            d.pop('Shares'),
            d.pop('Asset Mapping'))
        TradedAsset.from_dict(ret_obj, d)
        assert len(d) == 0, f'Extra attributes found: {list(d.keys())}'
        return ret_obj

    def to_table(self):
        """Returns a table representing this object.

        This function converts this object into a table
        suitable for pretty-printing.

        Returns: lakshmi.table.Table object representing the data in this
        class.
        """
        table = super().to_table()
        rows = table.list()
        rows.insert(0, ['Ticker:', f'{self._ticker}'])
        table.set_rows(rows)
        return table

    def cache_key(self):
        """Unique key used for caching return values."""
        return self._ticker

    @cache(365)  # Name changes are rare.
    def name(self):
        """Returns full name of this asset.

        This function pulls the name corresponding to the ticker symbol
        of this asset. The return value is cached for 365 days.

        Returns: A string representing the name of this asset.

        Raises: NonFoundError if the ticker is not found.
        """
        asset_name = self.yticker.info.get('longName') or \
            self.yticker.info.get('shortName') or \
            self.yticker.info.get('name')

        if asset_name is None:
            raise NotFoundError(
                f'Cannot retrieve ticker ("{self._ticker}") '
                'from Yahoo Finance')
        return asset_name

    def short_name(self):
        """Returns the short name (ticker) of this object."""
        return self._ticker

    @cache(1)
    def price(self):
        """Returns the market price of this asset.

        The return price is cached for a day.

        Returns: Price (float).

        Raises: NotFoundError if the ticker is not found.
        """
        if self.yticker.info.get('regularMarketPrice') is None:
            raise NotFoundError(
                f'Cannot retrieve ticker ("{self._ticker}") '
                'from Yahoo Finance')
        return self.yticker.info['regularMarketPrice']


class VanguardFund(TradedAsset, Cacheable):
    """An asset class representing Vanguard trust fund represented by a
    numeric ID."""

    def __init__(self, fund_id, shares, class2ratio):
        """
        Args:
            fund_id: Integer representing the Fund Id.
            shares: Number of shares of this fund.
            class2ratio: Dict of class_name -> ratio, where 0 < ratio <= 1.0
        """
        self._fund_id = fund_id
        super().__init__(shares, class2ratio)

    def to_dict(self):
        """Returns a dict representing this object."""
        d = {'Fund Id': self._fund_id,
             'Shares': self.shares(),
             'Asset Mapping': self.class2ratio}
        d.update(super().to_dict())
        return d

    @classmethod
    def from_dict(cls, d):
        """Returns a new object specified by dictionary d.

        This is reverse of to_dict.
        Args:
            d: A dictionary (usually the output of to_dict).

        Returns: A new VanguardFund object.

        Raises: AssertionError if d cannot be parsed correctly.
        """
        ret_obj = VanguardFund(
            d.pop('Fund Id'),
            d.pop('Shares'),
            d.pop('Asset Mapping'))
        TradedAsset.from_dict(ret_obj, d)
        assert len(d) == 0, f'Extra attributes found: {list(d.keys())}'
        return ret_obj

    def to_table(self):
        """Returns this asset as a lakshmi.table.Table object."""
        table = super().to_table()
        rows = table.list()
        rows.insert(0, ['Fund id:', f'{self._fund_id}'])
        table.set_rows(rows)
        return table

    def cache_key(self):
        """Unique key used for caching return values."""
        return str(self._fund_id)

    @cache(365)  # Name changes are very rare.
    def name(self):
        """Returns full name of this asset.

        This function returns the name of the fund corresponding to
        the fund id. The returned name is cached for 365 days.

        Returns: A string representing the name of this asset.

        Raises: AssertionError if the name cannot be fatched.
        """
        req = requests.get(
            f'https://api.vanguard.com/rs/ire/01/pe/fund/{self._fund_id}'
            '/profile.json',
            headers={'Referer': 'https://vanguard.com/'})
        req.raise_for_status()  # Raise if error
        return req.json()['fundProfile']['longName']

    def short_name(self):
        """Returns the short name (fund id, string) of this object."""
        return str(self._fund_id)

    @cache(1)
    def price(self):
        """Returns the market price of this asset.

        The return price is cached for a day.

        Returns: Price (float).

        Raises: AssertionError in case the price cannot be fetched.
        """
        req = requests.get(
            f'https://api.vanguard.com/rs/ire/01/pe/fund/{self._fund_id}'
            '/price.json',
            headers={'Referer': 'https://vanguard.com/'})
        req.raise_for_status()  # Raise if error
        return float(req.json()['currentPrice']
                     ['dailyPrice']['regular']['price'])


class _TreasuryBonds(Asset):
    """Class representing a collection of I or EE bonds."""
    class Bond(Cacheable):
        """A class representing individual I or EE Bond."""

        def __init__(self, series, issue_date, denom):
            """
            Args:
                series: Type of Bond, either 'I' or 'EE'.
                issue_date: String representing the issue month, in MM/YYYY
                format.
                denom: The denomination of this bond.
            """
            self.series = series
            self.issue_date = issue_date
            self.denom = denom
            self.redemption_date = datetime.datetime.now().strftime('%m/%Y')

        def cache_key(self):
            """Unique key used for caching return values."""
            return '{}_{}_{}'.format(
                self.series,
                self.issue_date.replace('/', '.'),
                self.redemption_date.replace('/', '.'))

        @cache(32)  # The value of a Bond doesn't change in a month.
        def _get_bond_info(self):
            """Returns the rate and value of a $1000 bond.

            Returns: A tuple representing the percentage rate and
            the current dollar value.
            """
            data = {
                'RedemptionDate': self.redemption_date,
                'Series': self.series,
                'Denomination': '1000',
                'IssueDate': self.issue_date,
                'btnAdd.x': 'CALCULATE'
            }

            req = requests.post(
                'http://www.treasurydirect.gov/BC/SBCPrice', data=data)
            req.raise_for_status()

            ret_vals = re.findall('\n<td>.*</td>', req.text)
            rate = re.sub('\n|<[^>]+>', '', ret_vals[6])
            value = float(re.sub('\n|\\$|,|<[^>]+>', '', ret_vals[7]))
            # EE Bonds returned are half the value (I guess TD website
            # assumes paper bonds)
            return rate, value * (2.0 if self.series == 'EE' else 1.0)

        def value(self):
            """Returns the current value of this bond."""
            unused_rate, value = self._get_bond_info()
            return value * (self.denom / 1000.0)

        def rate(self):
            """Returns the current percentage rate of this bond."""
            rate, unused_value = self._get_bond_info()
            return rate

        def as_list(self):
            """Returns this bond as a list.

            Returns: A list of Issue date, denomination, rate (as percentage)
            and current value.
            """
            rate, value = self._get_bond_info()
            value *= (self.denom / 1000.0)
            return [self.issue_date, self.denom, rate, value]

    def __init__(self, series, class2ratio):
        """
        Args:
            series: The type of the bonds, either 'I' or 'EE'.
            class2ratio: Dict of class_name -> ratio, where 0 < ratio <= 1.0
        """
        self._series = series
        super().__init__(class2ratio)
        self._bonds = []

    def to_dict(self):
        """Returns a dict representing this object."""
        d = {}
        d['Bonds'] = []
        for bond in self._bonds:
            d['Bonds'].append(
                {'Issue Date': bond.issue_date, 'Denomination': bond.denom})
        d.update(super().to_dict())
        return d

    def from_dict(self, d):
        """Returns a new object specified by dictionary d.

        This is reverse of to_dict.

        Args:
            d: A dictionary (usually the output of to_dict).

        Returns: A new _TreasuryBonds object.

        Raises: AssertionError if d cannot be parsed correctly.
        """
        for bond in d.pop('Bonds'):
            self.add_bond(bond.pop('Issue Date'), bond.pop('Denomination'))
            assert len(bond) == 0, ('Extra attributes found: '
                                    f'{list(bond.keys())}')
        Asset.from_dict(self, d)
        return self

    def bonds(self):
        """Returns all bonds as list.

        Returns: A lot of self.Bond objects.
        """
        return self._bonds

    def add_bond(self, issue_date, denom):
        """Adds a new bond to this asset.

        Args:
            issue_date: String representing the issue date (in MM/YYYY format)
            denom: The denomination of this bond.
        """
        self._bonds.append(self.Bond(self._series, issue_date, denom))
        return self

    def value(self):
        """Returns the current market value of all the bonds."""
        value = 0.0
        for bond in self._bonds:
            value += bond.value()
        return value

    def list_bonds(self):
        """Returns all bonds as a table.

        Returns: A lakshmi.table.Table contains all the bonds in this asset.
        The columns correspond to Issue Date, Denomination, Rate as
        percentage, and current market value.
        """
        table = Table(
            4,
            headers=['Issue Date', 'Denom', 'Rate', 'Value'],
            coltypes=['str', 'dollars', 'str', 'dollars'])
        for bond in self._bonds:
            table.add_row(bond.as_list())
        return table

    def string(self):
        """Returns this asset as string."""
        return (super().string() + '\n\nBonds:\n'
                + f'{self.list_bonds().string()}')

    def name(self):
        """Returns the name of this asset (either 'I Bonds' or 'EE Bonds')."""
        return f'{self._series} Bonds'

    def short_name(self):
        """Returns short name (same as name)."""
        return self.name()


class IBonds(_TreasuryBonds):
    """Class representing a collection of I Bonds."""
    def __init__(self, class2ratio):
        """
        Args:
            class2ratio: Dict of class_name -> ratio, where 0 < ratio <= 1.0
        """
        super().__init__('I', class2ratio)

    def to_dict(self):
        """Returns a dict representing this object."""
        d = {'Asset Mapping': self.class2ratio}
        d.update(super().to_dict())
        return d

    @classmethod
    def from_dict(cls, d):
        """Returns a new object specified by dictionary d."""
        ret_obj = IBonds(d.pop('Asset Mapping'))
        _TreasuryBonds.from_dict(ret_obj, d)
        assert len(d) == 0, f'Extra attributes found: {list(d.keys())}'
        return ret_obj


class EEBonds(_TreasuryBonds):
    def __init__(self, class2ratio):
        """
        Args:
            class2ratio: Dict of class_name -> ratio, where 0 < ratio <= 1.0
        """
        super().__init__('EE', class2ratio)

    def to_dict(self):
        """Returns a dict representing this object."""
        d = {'Asset Mapping': self.class2ratio}
        d.update(super().to_dict())
        return d

    @classmethod
    def from_dict(cls, d):
        """Returns a new object specified by dictionary d."""
        ret_obj = EEBonds(d.pop('Asset Mapping'))
        _TreasuryBonds.from_dict(ret_obj, d)
        assert len(d) == 0, f'Extra attributes found: {list(d.keys())}'
        return ret_obj


# A list of all the assets type (classes) defined in this module.
CLASSES = [ManualAsset, TickerAsset, VanguardFund, IBonds, EEBonds]



"""Classes to analyze the portfolio.

These classes are not meant to change the portfolio. They are used for
just analyzing the portfolio and output the results (if any).
"""

from abc import ABC, abstractmethod

from lakshmi.table import Table


# TODO(sarvjeets): This is not needed/used anywhere. Consider removing it.
class Analyzer(ABC):
    @abstractmethod
    def analyze(self, portfolio):
        pass


class TLH(Analyzer):
    """Tax loss harvesting opportunity analyzer.

    This triggers a message with details of tax lots that could be sold to
    harvest the losses.
    """

    def __init__(self, max_percentage, max_dollars=None):
        """Triggers if a lot has lost more than max_percentage or the total loss
        for a given asset exceeds max_dollars.

        Args:
            max_percentage: A float representing the max percentage loss after
            which a lot can be harvested.
            max_dollars: A float representing the max amount of loss (in
            dollars) for one or more lots (of the same security), beyond which
            we should recommend selling all the lots that are below their
            cost basis.
        """
        assert max_percentage > 0.0 and max_percentage < 1.0, (
            'max_percetage should be between 0% and 100% (exclusive).')
        self.max_percentage = max_percentage
        if max_dollars:
            assert max_dollars > 0, 'max_dollars should be positive.'
        self.max_dollars = max_dollars

    def _return_lots_to_sell(self, price, tax_lots):
        """Helper function that returns which lots to harvest.

        Args:
            price: The current price of the security.
            tax_lots: A list of lakshmi.assets.TaxLot (all belonging
            to the same security).

        Returns: A list of [lot date, loss in dollar, loss in percent]
        that could be loss harvested.
        """
        percent_lots = []
        negative_lots = []
        total_loss = 0.0

        for lot in tax_lots:
            loss = (lot.unit_cost - price) * lot.quantity
            loss_percent = (lot.unit_cost - price) / lot.unit_cost
            if loss > 0:
                negative_lots.append([lot.date, loss, loss_percent])
                total_loss += loss
            if loss_percent > self.max_percentage:
                percent_lots.append([lot.date, loss, loss_percent])

        if self.max_dollars is not None and total_loss > self.max_dollars:
            return negative_lots
        else:
            return percent_lots

    def analyze(self, portfolio):
        """Returns which lots can be tax loss harvested from a portfolio.

        Args:
            portfolio: A lakshmi.Portfolio object representing a portfolio.

        Returns: A lakshmi.table.Table object with tax lots to harvest with
        columns corespnding to 'Account', 'Asset', 'Date', 'Loss' and 'Loss%'.
        """
        ret_val = Table(
            5,
            headers=['Account', 'Asset', 'Date', 'Loss', 'Loss%'],
            coltypes=['str', 'str', 'str', 'dollars', 'percentage'])
        for account in portfolio.accounts():
            for asset in account.assets():
                if hasattr(asset, 'get_lots') and asset.get_lots():
                    for lot in self._return_lots_to_sell(
                            asset.price(), asset.get_lots()):
                        ret_val.add_row(
                            [account.name(), asset.short_name()] + lot)
        return ret_val


class BandRebalance(Analyzer):
    """Triggers if portfolio asset class targets are outside the bands
    specified. This considers an asset class outside bound if the absolute
    difference in percentage allocation is more than max_abs_percent different
    from target allocation or more than max_relative_percent different from the
    target allocation.

    A popular version is the 5/25 band based rebalancing rule. For more
    information, please see 5/25 on:
    https://www.bogleheads.org/wiki/Rebalancing
    """

    def __init__(self, max_abs_percent=0.05, max_relative_percent=0.25):
        """Constructor to set the bands. An asset class is considered outside
        the rebalance bands if the differnt between the desired and actual
        ratio of that asset class exceeds the lessor of max absolute or max
        relative percentage.

        Args:
            max_abs_percent: A float (ratio) representing the maximum absolute
            percent that an asset can deviate before it's considered outside
            the rebalancing band.
            max_relative_percent: A float (ratio) representing the maximum
            relative percent that an asset can deviate before it's considered
            outside the rebalancing band.
        """
        assert max_abs_percent > 0 and max_abs_percent < 1.0
        assert max_relative_percent > 0 and max_relative_percent < 1.0
        self.max_abs_percent = max_abs_percent
        self.max_relative_percent = max_relative_percent

    def analyze(self, portfolio):
        """Returns asset classes that are outside the rebalancing bands.

        Args:
            portfolio: A lakshmi.Portfolio object representing the portfolio
            that is being analyzed.

        Returns: A lakshmi.table.Table object with columns representing
        Asset Class, Actual % of asset class in portfolio, Desired % of asset
        class in the portfolio, Value in dollars of the asset class and the
        difference between Desired and actual dollar values.
        """
        aa = portfolio.asset_allocation(portfolio.asset_classes.leaves())
        headers = ['Class', 'Actual%', 'Desired%', 'Value', 'Difference']
        assert headers == aa.headers()
        ret_val = Table(
            5,
            headers,
            ['str', 'percentage', 'percentage', 'dollars', 'delta_dollars'])
        for row in aa.list():
            abs_percent = abs(row[1] - row[2])
            rel_percent = abs_percent / row[2] if row[2] != 0 else 0
            if (
                abs_percent >= self.max_abs_percent
                or rel_percent >= self.max_relative_percent
            ):
                ret_val.add_row(row)

        return ret_val



"""Common utils for Lakshmi."""


def format_money(x):
    """Formats input (money) to a string.

    For example, if x=5.238, the output is '$5.24'.

    Args:
        x: Float (non-negative) representing dollars.
    """
    return '${:,.2f}'.format(x)


def format_money_delta(x):
    """Formats input (money delta) into a string.

    For example, if x=-23.249m the output is '-$23.25'.

    Args:
        x: Float (postive or negative) representating dollars.
    """
    return '{}${:,.2f}'.format('-' if x < 0 else '+', abs(x))



from setuptools import find_packages, setup

from lakshmi.constants import NAME, VERSION

with open('README.md', 'r', encoding='utf-8') as fh:
    # Replace relative links to absolute links.
    long_description = (
        fh.read()
        .replace('](./', '](https://sarvjeets.github.io/lakshmi/')
        .replace('lak.md', 'lak.html'))

setup(
    name=NAME,
    version=VERSION,
    author='Sarvjeet Singh',
    author_email='sarvjeet@gmail.com',
    description=('Investing library and command-line interface '
                 'inspired by the Bogleheads philosophy'),
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/sarvjeets/lakshmi',
    platforms='any',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
    ],
    py_modules=['lakshmi'],
    packages=find_packages(exclude=['tests', 'tests.*']),
    include_package_data=True,
    install_requires=[
        'click~=8.0',
        'PyYAML>=5.4,<7.0',
        'requests~=2.25',
        'tabulate~=0.8',
        'yfinance~=0.1',
    ],
    entry_points={
        'console_scripts': [
            'lak = lakshmi.lak:lak',
        ],
    },
    test_suite='tests',
    python_requires='>=3.9.5',
)
